@@ -20,13 +20,15 @@
 #include <string.h>
 
 #include "icon-grid.h"
+#include "gtk-compat.h"
 
 /* Properties */
 enum {
   PROP_0,
   PROP_ORIENTATION,
   PROP_SPACING,
-  PROP_CONSTRAIN_WIDTH
+  PROP_CONSTRAIN_WIDTH,
+  PROP_ASPECT_WIDTH
   //PROP_FILL_WIDTH
 };
 
@@ -41,10 +43,10 @@ struct _PanelIconGrid
     gint spacing;				/* Desired spacing between grid elements */
     gint target_dimension;			/* Desired dimension perpendicular to orientation */
     gboolean constrain_width : 1;		/* True if width should be constrained by allocated space */
+    gboolean aspect_width : 1;			/* True if children should maintain aspect */
     gboolean fill_width : 1;			/* True if children should fill unused width */
     int rows;					/* Computed layout rows */
     int columns;				/* Computed layout columns */
-    int constrained_child_width;		/* Child width constrained by allocation */
     GdkWindow *event_window;			/* Event window if NO_WINDOW is set */
 };
 
@@ -53,7 +55,22 @@ struct _PanelIconGridClass
     GtkContainerClass parent_class;
 };
 
-static void panel_icon_grid_size_request(GtkWidget *widget, GtkRequisition *requisition);
+static void icon_grid_element_check_requisition(PanelIconGrid *ig,
+                                                GtkRequisition *requisition)
+{
+    if (ig->aspect_width && !ig->constrain_width &&
+        requisition->width > 1 && requisition->height > 1)
+    {
+        /* calculate width from aspect */
+        gdouble ratio = (gdouble)requisition->width / requisition->height;
+        requisition->width = ig->child_height * ratio;
+    }
+    else
+    {
+        requisition->width = ig->child_width;
+    }
+    requisition->height = ig->child_height;
+}
 
 /* Establish the widget placement of an icon grid. */
 static void panel_icon_grid_size_allocate(GtkWidget *widget,
@@ -66,21 +83,18 @@ static void panel_icon_grid_size_allocate(GtkWidget *widget,
     int child_height;
     GtkTextDirection direction;
     guint border;
-    int limit;
-    int x_initial;
     int x_delta;
-    int x, y;
+    guint next_coord;
+    guint x, y;
     GList *ige;
     GtkWidget *child;
 
     /* Apply given allocation */
     gtk_widget_set_allocation(widget, allocation);
     border = gtk_container_get_border_width(GTK_CONTAINER(widget));
-#if GTK_CHECK_VERSION(2, 20, 0)
+    child_allocation.width = MAX(allocation->width - border * 2, 0);
+    child_allocation.height = MAX(allocation->height - border * 2, 0);
     if (gtk_widget_get_realized(widget))
-#else
-    if (GTK_WIDGET_REALIZED(widget))
-#endif
     {
         if (!gtk_widget_get_has_window(widget))
         {
@@ -92,8 +106,6 @@ static void panel_icon_grid_size_allocate(GtkWidget *widget,
             child_allocation.x = 0;
             child_allocation.y = 0;
         }
-        child_allocation.width = MAX(allocation->width - border * 2, 0);
-        child_allocation.height = MAX(allocation->height - border * 2, 0);
         if (ig->event_window != NULL)
             gdk_window_move_resize(ig->event_window,
                                    child_allocation.x,
@@ -116,49 +128,91 @@ static void panel_icon_grid_size_allocate(GtkWidget *widget,
     child_width = ig->child_width;
     child_height = ig->child_height;
 
-    /* Calculate required size without borders */
-    panel_icon_grid_size_request(widget, &req);
-    req.width -= 2 * border;
-    req.height -= 2 * border;
+    /* FIXME: is there any sense to recheck rows and columns again?
+       GTK+ should have it done right before this call. */
 
     /* Get the constrained child geometry if the allocated geometry is insufficient.
      * All children are still the same size and share equally in the deficit. */
-    ig->constrained_child_width = ig->child_width;
-    if ((ig->columns != 0) && (ig->rows != 0) && (allocation->width > 1))
+    if ((ig->columns != 0) && (ig->rows != 0) && (child_allocation.width > 0))
     {
-        if (req.width > allocation->width && ig->constrain_width)
-            ig->constrained_child_width = child_width = (allocation->width + ig->spacing - 2 * border) / ig->columns - ig->spacing;
-        if (ig->orientation == GTK_ORIENTATION_HORIZONTAL && req.height < allocation->height)
-            child_height = (allocation->height + ig->spacing - 2 * border) / ig->rows - ig->spacing;
+        if (ig->constrain_width &&
+            (x_delta = (child_allocation.width + ig->spacing) / ig->columns - ig->spacing) < child_width)
+            child_width = MAX(2, x_delta);
+        /* fill vertical space evenly in horisontal orientation */
+        if (ig->orientation == GTK_ORIENTATION_HORIZONTAL &&
+            (x_delta = (child_allocation.height + ig->spacing) / ig->rows - ig->spacing) > child_height)
+            child_height = MAX(2, x_delta);
     }
 
     /* Initialize parameters to control repositioning each visible child. */
     direction = gtk_widget_get_direction(widget);
-    limit = border + ((ig->orientation == GTK_ORIENTATION_HORIZONTAL)
-        ?  (ig->rows * (child_height + ig->spacing))
-        :  (ig->columns * (child_width + ig->spacing)));
-    x_initial = ((direction == GTK_TEXT_DIR_RTL)
-        ? allocation->width - child_width - border
-        : border);
-    x_delta = child_width + ig->spacing;
-    if (direction == GTK_TEXT_DIR_RTL) x_delta = - x_delta;
+    x = (direction == GTK_TEXT_DIR_RTL) ? allocation->width - border : border;
+    y = border;
+    x_delta = 0;
+    next_coord = border;
 
     /* Reposition each visible child. */
-    x = x_initial;
-    y = border;
     for (ige = ig->children; ige != NULL; ige = ige->next)
     {
         child = ige->data;
         if (gtk_widget_get_visible(child))
         {
             /* Do necessary operations on the child. */
-            gtk_widget_size_request(child, &req);
-            child_allocation.x = x;
-            child_allocation.y = y;
-            child_allocation.width = child_width;
+            gtk_widget_get_child_requisition(child, &req);
+            icon_grid_element_check_requisition(ig, &req);
+            child_allocation.width = MIN(req.width, child_width);
             child_allocation.height = MIN(req.height, child_height);
+
+            /* Check this grid position */
+            if (ig->orientation == GTK_ORIENTATION_HORIZONTAL)
+            {
+                y = next_coord;
+                if (y + child_height > allocation->height - border && y > border)
+                {
+                    y = border;
+                    if (direction == GTK_TEXT_DIR_RTL)
+                        x -= (x_delta + ig->spacing);
+                    else
+                        x += (x_delta + ig->spacing);
+                    x_delta = 0;
+                    // FIXME: if fill_width and rows = 1 then allocate whole column
+                }
+                next_coord = y + child_height + ig->spacing;
+                x_delta = MAX(x_delta, child_allocation.width);
+            }
+            else
+            {
+                // FIXME: if fill_width then use aspect to check delta
+                if (direction == GTK_TEXT_DIR_RTL)
+                {
+                    next_coord = x - child_allocation.width;
+                    if (x < allocation->width - border)
+                    {
+                        next_coord -= ig->spacing;
+                        if (next_coord < border)
+                        {
+                            next_coord = allocation->width - border;
+                            y += child_height + ig->spacing;
+                        }
+                    }
+                    x = next_coord;
+                }
+                else
+                {
+                    x = next_coord;
+                    if (x + child_allocation.width > allocation->width - border && x > border)
+                    {
+                        x = border;
+                        y += child_height + ig->spacing;
+                    }
+                    next_coord = x + child_allocation.width + ig->spacing;
+                }
+            }
+            child_allocation.x = x;
             if (req.height < child_height - 1)
-                child_allocation.y += (child_height - req.height) / 2;
+                y += (child_height - req.height) / 2;
+            child_allocation.y = y;
+
             if (!gtk_widget_get_has_window (widget))
             {
                 child_allocation.x += allocation->x;
@@ -166,28 +220,6 @@ static void panel_icon_grid_size_allocate(GtkWidget *widget,
             }
             // FIXME: if fill_width and rows > 1 then delay allocation
             gtk_widget_size_allocate(child, &child_allocation);
-
-            /* Advance to the next grid position. */
-            if (ig->orientation == GTK_ORIENTATION_HORIZONTAL)
-            {
-                y += child_height + ig->spacing;
-                if (y >= limit)
-                {
-                    y = border;
-                    x += x_delta;
-                    // FIXME: if fill_width and rows = 1 then allocate whole column
-                }
-            }
-            else
-            {
-                // FIXME: if fill_width then use aspect to check delta
-                x += x_delta;
-                if ((direction == GTK_TEXT_DIR_RTL) ? (x <= 0) : (x >= limit))
-                {
-                    x = x_initial;
-                    y += child_height + ig->spacing;
-                }
-            }
         }
     }
 }
@@ -197,28 +229,47 @@ static void panel_icon_grid_size_request(GtkWidget *widget,
                                          GtkRequisition *requisition)
 {
     PanelIconGrid *ig = PANEL_ICON_GRID(widget);
-    int visible_children = 0;
     GList *ige;
     int target_dimension = ig->target_dimension;
     guint border = gtk_container_get_border_width(GTK_CONTAINER(widget));
     gint old_rows = ig->rows;
     gint old_columns = ig->columns;
+    gint row = 0, w = 0;
+    GtkRequisition child_requisition;
 
-    /* Count visible children. */
-    for (ige = ig->children; ige != NULL; ige = ige->next)
-        if (gtk_widget_get_visible(ige->data))
-            visible_children += 1;
-
+    requisition->width = 0;
+    requisition->height = 0;
+    ig->rows = 0;
+    ig->columns = 0;
     if (ig->orientation == GTK_ORIENTATION_HORIZONTAL)
     {
         /* In horizontal orientation, fit as many rows into the available height as possible.
          * Then allocate as many columns as necessary.  Guard against zerodivides. */
-        ig->rows = 0;
         if ((ig->child_height + ig->spacing) != 0)
             ig->rows = (target_dimension + ig->spacing - border * 2) / (ig->child_height + ig->spacing);
         if (ig->rows == 0)
             ig->rows = 1;
-        ig->columns = (visible_children + (ig->rows - 1)) / ig->rows;
+        /* Count visible children and columns. */
+        for (ige = ig->children; ige != NULL; ige = ige->next)
+            if (gtk_widget_get_visible(ige->data))
+            {
+                gtk_widget_size_request(ige->data, &child_requisition);
+                icon_grid_element_check_requisition(ig, &child_requisition);
+                if (row == 0)
+                    ig->columns++;
+                w = MAX(w, child_requisition.width);
+                row++;
+                if (row == ig->rows)
+                {
+                    row = 0;
+                    if (requisition->width > 0)
+                        requisition->width += ig->spacing;
+                    requisition->width += w;
+                    row = w = 0;
+                }
+            }
+        if (row > 0)
+            requisition->width += w;
         /* if ((ig->columns == 1) && (ig->rows > visible_children))
             ig->rows = visible_children; */
     }
@@ -226,46 +277,79 @@ static void panel_icon_grid_size_request(GtkWidget *widget,
     {
         /* In vertical orientation, fit as many columns into the available width as possible.
          * Then allocate as many rows as necessary.  Guard against zerodivides. */
-        ig->columns = 0;
         if ((ig->child_width + ig->spacing) != 0)
             ig->columns = (target_dimension + ig->spacing - border * 2) / (ig->child_width + ig->spacing);
         if (ig->columns == 0)
             ig->columns = 1;
-        ig->rows = (visible_children + (ig->columns - 1)) / ig->columns;
-        if ((ig->rows == 1) && (ig->columns > visible_children))
-            ig->columns = visible_children;
+        /* Count visible children and rows. */
+        for (ige = ig->children; ige != NULL; ige = ige->next)
+            if (gtk_widget_get_visible(ige->data))
+            {
+                gtk_widget_size_request(ige->data, &child_requisition);
+                icon_grid_element_check_requisition(ig, &child_requisition);
+                if (w > 0 && w + child_requisition.width > target_dimension)
+                {
+                    w = 0;
+                    ig->rows++;
+                }
+                w += child_requisition.width;
+                requisition->width = MAX(requisition->width, w);
+            }
+        if (w > 0)
+            ig->rows++;
     }
 
     /* Compute the requisition. */
     if ((ig->columns == 0) || (ig->rows == 0))
-    {
-        requisition->width = 0;
         requisition->height = 0;
-    }
     else
-    {
-        int column_spaces = ig->columns - 1;
-        int row_spaces = ig->rows - 1;
-        if (column_spaces < 0) column_spaces = 0;
-        if (row_spaces < 0) row_spaces = 0;
-        requisition->width = ig->child_width * ig->columns + column_spaces * ig->spacing + 2 * border;
-        requisition->height = ig->child_height * ig->rows + row_spaces * ig->spacing + 2 * border;
-    }
+        requisition->height = (ig->child_height + ig->spacing) * ig->rows - ig->spacing + 2 * border;
+
     if (ig->rows != old_rows || ig->columns != old_columns)
         gtk_widget_queue_resize(widget);
 }
 
-/* Handler for "size-request" event on the icon grid element. */
-static void icon_grid_element_size_request(GtkWidget * widget, GtkRequisition * requisition, PanelIconGrid * ig)
+#if GTK_CHECK_VERSION(3, 0, 0)
+static void panel_icon_grid_get_preferred_width(GtkWidget *widget,
+                                                gint *minimal_width,
+                                                gint *natural_width)
 {
-    /* This is our opportunity to request space for the element. */
-    // FIXME: if fill_width then calculate width from aspect
-    requisition->width = ig->child_width;
-    if ((ig->constrain_width) && (ig->constrained_child_width > 1))
-        requisition->width = ig->constrained_child_width;
-    requisition->height = ig->child_height;
+    PanelIconGrid *ig = PANEL_ICON_GRID(widget);
+    GtkRequisition requisition;
+
+    if (ig->orientation == GTK_ORIENTATION_VERTICAL)
+    {
+        if (minimal_width)
+            *minimal_width = MIN(ig->target_dimension, ig->child_width);
+        if (natural_width)
+            *natural_width = ig->target_dimension;
+        return;
+    }
+    panel_icon_grid_size_request(widget, &requisition);
+    if (minimal_width)
+        *minimal_width = requisition.width;
+    if (natural_width)
+        *natural_width = requisition.width;
 }
 
+static void panel_icon_grid_get_preferred_height(GtkWidget *widget,
+                                                 gint *minimal_height,
+                                                 gint *natural_height)
+{
+    PanelIconGrid *ig = PANEL_ICON_GRID(widget);
+    GtkRequisition requisition;
+
+    if (ig->orientation == GTK_ORIENTATION_HORIZONTAL)
+        requisition.height = ig->target_dimension;
+    else
+        panel_icon_grid_size_request(widget, &requisition);
+    if (minimal_height)
+        *minimal_height = requisition.height;
+    if (natural_height)
+        *natural_height = requisition.height;
+}
+#endif
+
 /* Add an icon grid element and establish its initial visibility. */
 static void panel_icon_grid_add(GtkContainer *container, GtkWidget *widget)
 {
@@ -275,8 +359,6 @@ static void panel_icon_grid_add(GtkContainer *container, GtkWidget *widget)
     ig->children = g_list_append(ig->children, widget);
 
     /* Add the widget to the layout container. */
-    g_signal_connect(G_OBJECT(widget), "size-request",
-                     G_CALLBACK(icon_grid_element_size_request), container);
     gtk_widget_set_parent(widget, GTK_WIDGET(container));
 //    gtk_widget_queue_resize(GTK_WIDGET(container));
 }
@@ -293,6 +375,17 @@ void panel_icon_grid_set_constrain_width(PanelIconGrid * ig, gboolean constrain_
     gtk_widget_queue_resize(GTK_WIDGET(ig));
 }
 
+void panel_icon_grid_set_aspect_width(PanelIconGrid * ig, gboolean aspect_width)
+{
+    g_return_if_fail(PANEL_IS_ICON_GRID(ig));
+
+    if ((!ig->aspect_width && !aspect_width) || (ig->aspect_width && aspect_width))
+        return;
+
+    ig->aspect_width = !!aspect_width;
+    gtk_widget_queue_resize(GTK_WIDGET(ig));
+}
+
 /* void panel_icon_grid_set_fill_width(PanelIconGrid * ig, gboolean fill_width)
 {
     g_return_if_fail(PANEL_IS_ICON_GRID(ig));
@@ -319,9 +412,6 @@ static void panel_icon_grid_remove(GtkContainer *container, GtkWidget *widget)
             gboolean was_visible = gtk_widget_get_visible(widget);
 
             /* The child is found.  Remove from child list and layout container. */
-            g_signal_handlers_disconnect_by_func(widget,
-                                                 icon_grid_element_size_request,
-                                                 container);
             gtk_widget_unparent (widget);
             ig->children = g_list_remove_link(ig->children, children);
             g_list_free(children);
@@ -398,7 +488,6 @@ void panel_icon_grid_set_geometry(PanelIconGrid * ig,
 
     ig->orientation = orientation;
     ig->child_width = child_width;
-    ig->constrained_child_width = child_width;
     ig->child_height = child_height;
     ig->spacing = spacing;
     ig->target_dimension = target_dimension;
@@ -437,6 +526,9 @@ static void panel_icon_grid_set_property(GObject *object, guint prop_id,
     case PROP_CONSTRAIN_WIDTH:
         panel_icon_grid_set_constrain_width(ig, g_value_get_boolean(value));
         break;
+    case PROP_ASPECT_WIDTH:
+        panel_icon_grid_set_aspect_width(ig, g_value_get_boolean(value));
+        break;
     /* case PROP_FILL_WIDTH:
         panel_icon_grid_set_fill_width(ig, g_value_get_boolean(value));
         break; */
@@ -462,6 +554,9 @@ static void panel_icon_grid_get_property(GObject *object, guint prop_id,
     case PROP_CONSTRAIN_WIDTH:
         g_value_set_boolean(value, ig->constrain_width);
         break;
+    case PROP_ASPECT_WIDTH:
+        g_value_set_boolean(value, ig->aspect_width);
+        break;
     /* case PROP_FILL_WIDTH:
         g_value_set_boolean(value, ig->fill_width);
         break; */
@@ -507,10 +602,16 @@ static void panel_icon_grid_realize(GtkWidget *widget)
     if (visible_window)
     {
         attributes.visual = gtk_widget_get_visual(widget);
+#if !GTK_CHECK_VERSION(3, 0, 0)
         attributes.colormap = gtk_widget_get_colormap(widget);
+#endif
         attributes.wclass = GDK_INPUT_OUTPUT;
 
+#if GTK_CHECK_VERSION(3, 0, 0)
+        attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL;
+#else
         attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
+#endif
 
         window = gdk_window_new(gtk_widget_get_parent_window(widget),
                                 &attributes, attributes_mask);
@@ -569,19 +670,33 @@ static void panel_icon_grid_unmap(GtkWidget *widget)
     GTK_WIDGET_CLASS(panel_icon_grid_parent_class)->unmap(widget);
 }
 
+#if GTK_CHECK_VERSION(3, 0, 0)
+static gboolean panel_icon_grid_draw(GtkWidget *widget, cairo_t *cr)
+#else
 static gboolean panel_icon_grid_expose(GtkWidget *widget, GdkEventExpose *event)
+#endif
 {
     if (gtk_widget_is_drawable(widget))
     {
         if (gtk_widget_get_has_window(widget) &&
             !gtk_widget_get_app_paintable(widget))
+#if GTK_CHECK_VERSION(3, 0, 0)
+            gtk_render_background(gtk_widget_get_style_context(widget), cr, 0, 0,
+                                  gtk_widget_get_allocated_width(widget),
+                                  gtk_widget_get_allocated_height(widget));
+#else
             gtk_paint_flat_box(gtk_widget_get_style(widget),
                                gtk_widget_get_window(widget),
                                gtk_widget_get_state(widget), GTK_SHADOW_NONE,
                                &event->area, widget, "panelicongrid",
                                0, 0, -1, -1);
+#endif
 
+#if GTK_CHECK_VERSION(3, 0, 0)
+        GTK_WIDGET_CLASS(panel_icon_grid_parent_class)->draw(widget, cr);
+#else
         GTK_WIDGET_CLASS(panel_icon_grid_parent_class)->expose_event(widget, event);
+#endif
     }
     return FALSE;
 }
@@ -617,13 +732,22 @@ static void panel_icon_grid_class_init(PanelIconGridClass *class)
     object_class->set_property = panel_icon_grid_set_property;
     object_class->get_property = panel_icon_grid_get_property;
 
+#if GTK_CHECK_VERSION(3, 0, 0)
+    widget_class->get_preferred_width = panel_icon_grid_get_preferred_width;
+    widget_class->get_preferred_height = panel_icon_grid_get_preferred_height;
+#else
     widget_class->size_request = panel_icon_grid_size_request;
+#endif
     widget_class->size_allocate = panel_icon_grid_size_allocate;
     widget_class->realize = panel_icon_grid_realize;
     widget_class->unrealize = panel_icon_grid_unrealize;
     widget_class->map = panel_icon_grid_map;
     widget_class->unmap = panel_icon_grid_unmap;
+#if GTK_CHECK_VERSION(3, 0, 0)
+    widget_class->draw = panel_icon_grid_draw;
+#else
     widget_class->expose_event = panel_icon_grid_expose;
+#endif
 
     container_class->add = panel_icon_grid_add;
     container_class->remove = panel_icon_grid_remove;
@@ -648,6 +772,12 @@ static void panel_icon_grid_class_init(PanelIconGridClass *class)
                                                          "Constrain width",
                                                          "Whether to constrain width by allocated space",
                                                          FALSE, G_PARAM_READWRITE));
+    g_object_class_install_property(object_class,
+                                    PROP_ASPECT_WIDTH,
+                                    g_param_spec_boolean("aspect-width",
+                                                         "Maintain children aspect",
+                                                         "Whether to set children width to maintain their aspect",
+                                                         FALSE, G_PARAM_READWRITE));
 }
 
 static void panel_icon_grid_init(PanelIconGrid *ig)
@@ -667,13 +797,13 @@ GtkWidget * panel_icon_grid_new(
     /* Create a structure representing the icon grid and collect the parameters. */
     PanelIconGrid * ig = g_object_new(PANEL_TYPE_ICON_GRID,
                                       "orientation", orientation,
-                                      "spacing", spacing, NULL);
+                                      "spacing", spacing,
+                                      "border-width", border,
+                                      NULL);
 
     ig->child_width = child_width;
-    ig->constrained_child_width = child_width;
     ig->child_height = child_height;
     ig->target_dimension = target_dimension;
-    gtk_container_set_border_width(GTK_CONTAINER(ig), border);
 
     return (GtkWidget *)ig;
 }
